# User interface
inputDataUploadUI <- function(id)
{
  ns <- NS(id)

  tagList(
    box(
      width = 12,
      title = "Specify input data file",
      solidHeader = TRUE,
      status = "primary",
      fileInput(ns("fileInput"), label = h3("File input"))),
    uiOutput(ns("origDataDetails")),
    uiOutput(ns("attrMapping"))
  )
}

# Server logic
inputDataUpload <- function(input, output, session)
{
  # Get namespace
  ns <- session$ns

  vals <- reactiveValues(attrMapping = list(),
                         attrMappingStatus = list(),
                         attrMappingValid = FALSE,
                         origDataAttrs = c(),
                         inputData = NULL)

  # EVENT: Input data file name changed
  originalData <- eventReactive(input[["fileInput"]], {
    # Get reference to file input
    fileInput <- input$fileInput

    # Validate
    validate(need(fileInput, message = FALSE))

    # Get input data file name and type
    fileType <- tolower(tools::file_ext(fileInput$name))
    fileName <- fileInput$datapath

    # Read input data
    originalData <- ReadDataFile(fileName, fileType)

    vals$origDataAttrs <- colnames(originalData)
    vals$attrMapping <- GetPreliminaryAttributesMapping(originalData)

    return(originalData)
  }, ignoreNULL = TRUE)

  # Output original data details when they have changed
  output[["origDataDetails"]] <- renderUI({
    # Respond to originalData change
    originalData <- originalData()

    isolate({
      # Get reference to file input
      fileInput <- input$fileInput

      # Get input data details
      inputDataDetails <-
        div(
          p(strong("Name:"), fileInput$name),
          p(strong("Size:"), FormatObjectSize(fileInput$size)),
          p(strong("Type:"), fileInput$type),
          p(strong("Number of records:"), nrow(originalData)),
          p(strong("Column names:"), tags$br(), paste(colnames(originalData), collapse = ", "))
        )

      # Wrap in a box
      inputDataDetails <- box(
        width = 12,
        title = "Input data basic details",
        solidHeader = TRUE,
        status = "warning",
        inputDataDetails)

      return(inputDataDetails)
    })
  })

  # Generate attributes mapping
  output[["attrMapping"]] <- renderUI({
    # Respond to originalData change
    originalData()

    isolate({

      ns <- session$ns

      mappingStatusMsg <- uiOutput(ns("attrMappingStatus"))
      descrText <- p("Please, provide mapping between attributes used internally by the tool (column \"Tool\") and data dimensions (column \"Data\").")
      headers <- fluidRow(
        column(3, h3("Tool")),
        column(9, h3("Data"))
      )

      attributesMappingRowsUI <- lapply(names(vals$attrMapping), function(analysisAttr) {
        selectOptions <- list(placeholder = paste("Select", analysisAttr, "attribute"))
        if (is.null(vals$attrMapping[[analysisAttr]])) {
          selectOptions <- list(
            placeholder = paste("Select", analysisAttr, "attribute"),
            onInitialize = I('function() { this.setValue(""); }'))
        }

        selectId <- paste0("attr_", analysisAttr)

        widget <- fluidRow(
          # Selection attribute
          column(3, h4(analysisAttr)),
          # Selection data dimension
          column(9, selectizeInput(ns(selectId),
                                   label = NULL,
                                   choices = vals$origDataAttrs,
                                   selected = vals$attrMapping[[analysisAttr]],
                                   options = selectOptions))
        )

        # EVENT: Selected item changed
        observeEvent(input[[selectId]], {
          vals$attrMapping[[analysisAttr]] <- input[[selectId]]
        })

        return(widget)
      })

      # Wrap in a box
      attributesMappingUI <- box(
        width = 12,
        title = "Attributes mapping",
        solidHeader = TRUE,
        status = "primary",
        tagList(descrText,
                mappingStatusMsg,
                headers,
                attributesMappingRowsUI))

      return(attributesMappingUI)
    })
  })

  observe({
    attrMappingStatus <- GetAttrMappingStatus(vals$attrMapping)
    vals$attrMappingValid <- attrMappingStatus$Valid
    vals$attrMappingStatus <- attrMappingStatus[c("NonMapped", "MultipleMapped")]
  })

  output[["attrMappingStatus"]] <- renderUI({
    nonMappedAnalysisAttrs <- vals$attrMappingStatus$NonMapped
    multipleMappedDataAttrs <- vals$attrMappingStatus$MultipleMapped
    attrMappingValid <- vals$attrMappingValid

    isolate({
      header <- NULL
      textColor <- NULL
      if (attrMappingValid) {
        textColor <- "green"
        header <- p("Mapping is valid.", style = paste("color:", textColor))
      } else {
        textColor <- "red"
        header <- p("Mapping is not valid. Please, correct the following issues:", style = paste("color:", textColor))
      }

      nonMappedStatus <- NULL
      if (length(nonMappedAnalysisAttrs) > 0) {
        nonMappedStatus <- tags$li(
          "Tool attributes not mapped:",
          tags$ul(
            lapply(nonMappedAnalysisAttrs, function(analysisAttr) tags$li(analysisAttr))
          )
        )
      }

      multiMappedStatus <- NULL
      if (length(multipleMappedDataAttrs) > 0) {
        multiMappedStatus <- tags$li(
          "Data attributes mapped multiple times:",
          tags$ul(
            lapply(names(multipleMappedDataAttrs), function(dataAttr) {
              analysisAttrs <- multipleMappedDataAttrs[[dataAttr]]
              tags$li(paste(AddQuoteMarks(dataAttr),
                            "mapped to",
                            paste(AddQuoteMarks(analysisAttrs), collapse = ", ")))
            })
          )
        )
      }

      attrMappingStatus <- tagList(header,
                                   tags$ol(nonMappedStatus,
                                           multiMappedStatus,
                                           style = paste("color:", textColor)))

      return(attrMappingStatus)
    })
  })

  inputData <- eventReactive(vals$attrMapping, {
    intermediateData <- ApplyAttributesMapping(originalData(), vals$attrMapping)
    PreProcessInputData(intermediateData)
  })

  return(list(originalData = originalData,
              inputData = inputData))
}
